; File:		updi-test-tn1614\main.asm
; Device:	ATtiny1614
; Created:	09.06.2024 21:52:27
; Version:	2024-06-23
; Author:	Johannes Fechner

.nolist
.include <tn1614def.inc>
.include "macros.asm"
.list

; == Hardware definitions ==
; === CPU clock frequency ===
.equ F_CPU = 20_000_000 ; Hz (internal HF oscillator w/o prescaling).

; === Hardware connections ===
; USART0.TxD output:
.equ TXD_VDIR = VPORTB_DIR
.equ TXD_bp = 2
; Target UPDI (connected via 1kOhm "anxiety" resistor to UPDI of target device):
.equ TARGET_UPDI_INTFLAGS = PORTA_INTFLAGS
.equ TARGET_UPDI_PINCTRL = PORTA_PIN1CTRL
.equ TARGET_UPDI_PORT_VECT = PORTA_PORT_vect
.equ TARGET_UPDI_VDIR = VPORTA_DIR
.equ TARGET_UPDI_VIN = VPORTA_IN
.equ TARGET_UPDI_VOUT = VPORTA_OUT
.equ TARGET_UPDI_bp = 1 ; PA1.
; Target UPDI Rx LED:
.equ LED_UPDI_RX_VDIR = VPORTA_DIR
.equ LED_UPDI_RX_VIN = VPORTA_IN
.equ LED_UPDI_RX_VOUT = VPORTA_OUT
.equ LED_UPDI_RX_bp = 2 ; PA2.
; Target UPDI Tx LED:
.equ LED_UPDI_TX_VDIR = VPORTA_DIR
.equ LED_UPDI_TX_VOUT = VPORTA_OUT
.equ LED_UPDI_TX_bp = 3 ; PA3.

; == PC UART definitions ==
.equ UART_BPS = 115_200 ; bits/s

; == UPDI protocol definitions ==
.equ UPDI_BPS = 50_000		; UPDI UART baud rate in bits/s.
.equ UPDI_UART_BUFLEN = 16	; UPDI UART tx/rx buffer length in bytes.

; == Register name definitions ==
.def zero = r2		; Contains always 0x00.

; === Temporary registers ===
.def tmp0 = r16
.def tmp1 = r17
.def tmp2 = r18
.def tmp3 = r19
.def tmpL = r24
.def tmpH = r25

; === Main registers ===
.def updiFlags = r3				; UPDI status bits and flags.
.equ UPDI_FLAGS_BUSY_bp = 0		; UPDI transmission or reception in progress.
.equ UPDI_FLAGS_TXC_bp = 1		; UPDI transmission complete.
.equ UPDI_FLAGS_RXC_bp = 2		; UPDI Frame Received flag.
.equ UPDI_FLAGS_TIMEOUT_bp = 3	; Timeout occurred while waiting for reception.
.equ UPDI_FLAGS_BUFOVF_bp = 4	; Buffer overflow while receiving.

; == RAM usage ==
.dseg
.org SRAM_START

updiUartBitCntr_ram:		.byte   1
; == 0: during start bit.
; == 1: during data LSb.
; ...
; == 8: during data MSb.
; == 9: during (even) parity bit.
; ==10: during 1st stop bit.
; ==11: during 2nd stop bit.
updiUartBitMask_ram:		.byte   1
updiUartDataBuf_ram:		.byte   UPDI_UART_BUFLEN
updiUartDataCount_ram:		.byte	1
updiUartDataIndex_ram:		.byte	1
updiUartParityCntr_ram:		.byte   1
updiUartTimeoutCntr_ram:	.byte   1

writeFuseAddrL_ram:			.byte   1
writeFuseAddrH_ram:			.byte   1
writeFuseData_ram:			.byte   1

.cseg
.org 0
	rjmp	reset

; == Interrupt vectors ==
.org TARGET_UPDI_PORT_VECT
	rjmp	updi_fallingEdgeIsr
.org TCA0_OVF_vect
	rjmp	updi_timerIsr
	
; == Zero-terminated strings ==
.org INT_VECTORS_SIZE
szMenu: .db "(1) Read SIB", '\r', '\n',\
	"(2) Read UPDI register", '\r', '\n',\
	"(3) Read byte from UPDI address space", '\r', '\n',\
	"(4) Chip erase", '\r', '\n',\
	"(5) Write fuse byte", '\r', '\n', 0, 0
szAddress: .db "Address: ", 0
szData: .db " Data: ", 0
szLdcs: .db "LDCS returned: ", 0
szLds: .db "LDS returned: ", 0, 0
szConfirm: .db " Confirm (y/n): ", 0, 0
szEnteringCEK: .db "Entering Chip Erase key...", '\r', '\n', 0, 0
szEnteringNVMPK: .db "Entering NVM Programming key...", '\r', '\n', 0
szAsiKeyStatus: .db "ASI KEY STATUS: ", 0, 0
szReset: .db "Issuing System Reset...", '\r', '\n', 0
szAsiSysStatus: .db "ASI SYS STATUS: ", 0, 0
szNvmctrlStatus: .db "NVMCTRL_STATUS: ", 0, 0
szNvmVersion: .db "NVM version number: ", 0, 0
szSendingSts0: .db "Sending STS no. 1/4 ... ", 0, 0
szSendingSts1: .db "Sending STS no. 2/4 ... ", 0, 0
szSendingSts2: .db "Sending STS no. 3/4 ... ", 0, 0
szSendingSts3: .db "Sending STS no. 4/4 ... ", 0, 0
szOK: .db "OK", '\r', '\n', 0, 0
szErrInvAnswer: .db "Error: Invalid answer.", '\r', '\n', '\r', '\n', 0, 0
szErrBufOvf: .db "Error: Buffer overflow while receiving.", '\r', '\n', '\r', '\n', 0
szErrTimeout: .db "Error: Timeout while waiting for UPDI answer.", '\r', '\n', '\r', '\n', 0
szErrKey: .db "Error: ASI Key submitting failed.", '\r', '\n', '\r', '\n', 0
szErrErasefail: .db"Error: ERASEFAIL==1 - Chip Erase failed.", '\r', '\n', '\r', '\n', 0, 0
szLineBreak: .db '\r', '\n', 0, 0

; == External routines ==
.include "updi.asm"

; == Execution entry point after reset ==
reset:
; == Set clock ==
; Disable clock prescaler:
	ldi		tmp0, CPU_CCP_IOREG_gc
	xout	CPU_CCP, tmp0
	ldi		tmp0, 0
	xout	CLKCTRL_MCLKCTRLB, tmp0

; == Initialize registers ==
	clr		zero
	clr		updiFlags

; == Initialize UART ==
.equ BAUD_REG_VAL = ROUND(4.0*F_CPU/(1.0*UART_BPS))
	ldi		tmp0, low(BAUD_REG_VAL)
	xout	USART0_BAUDL, tmp0
	ldi		tmp0, high(BAUD_REG_VAL)
	xout	USART0_BAUDH, tmp0
; USART0_CTRLC = $03 (reset value): asynchronous, 8N1.
	ldi		tmp0, (1<<USART_TXEN_bp) | (1<<USART_RXEN_bp)
	xout	USART0_CTRLB, tmp0

; == Initialize TCA0 for UPDI Soft-UART ==
; Enable overflow interrupt:
	ldi		tmp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTCTRL, tmp0

; == Initialize I/O ==
; Configure TxD pin as output:
	sbi		TXD_VDIR, TXD_bp
; Configure Rx/Tx LEDs as outputs:
	sbi		LED_UPDI_RX_VDIR, LED_UPDI_RX_bp
	sbi		LED_UPDI_TX_VDIR, LED_UPDI_TX_bp
; Enable pull-up at Target UPDI pin:
	ldi		tmp0, PORT_PULLUPEN_bm
	xout	TARGET_UPDI_PINCTRL, tmp0

; == Enable interrupts ==
	sei
	
loop:
; Print menu to PC UART:
	ldiz	2*szMenu
	rcall	uart_putSz
; Wait for PC UART reception:
	rcall	uart_getChar
	cpi		tmp0, '1'
	brne	loop_2
	rjmp	loop_readSib
loop_2:
	cpi		tmp0, '2'
	brne	loop_3
	rjmp	loop_readUpdiReg
loop_3:
	cpi		tmp0, '3'
	brne	loop_4
	rjmp	loop_readByte
loop_4:
	cpi		tmp0, '4'
	brne	loop_5
	rjmp	loop_chipErase
loop_5:
	cpi		tmp0, '5'
	brne	loop_default
	rjmp	loop_writeFuse
loop_default:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop

loop_readSib:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_readSib_waitTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_readSib_waitTx
; Wait until all SIB bytes were received:
loop_readSib_waitRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_readSib_waitRx
; updiUartDataBuf_ram now contains the received bytes.
	clr		tmp1
    ldiz	updiUartDataBuf_ram
loop_readSib_print:
	ld		tmp0, Z+
    rcall   uart_putChar
	inc		tmp1
	cpi		tmp1, 16
	brlo	loop_readSib_print
	ldiz	2*szLineBreak
	rcall	uart_putSz
    ldiz    2*szLineBreak
    rcall   uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_readUpdiReg:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Get address from user:
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send LDCS instruction:
	andi	tmp0, 0x0F
	rcall	sendLdcs
; Wait for answer bytes from Target UPDI:
loop_readUpdiReg_wait:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_readUpdiReg_wait
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
	ldiz	2*szLdcs
	rcall	uart_putSz
    rcall   uart_putHexByte
    ldiz    2*szLineBreak
    rcall   uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_readByte:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Get address from user:
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte	; Address MSB.
	mov		YH, tmp0
	rcall	uart_getHexByte	; Address LSB.
	mov		YL, tmp0
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command, to keep UPDI alive and get the NVM version:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_readByte_waitSibTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_readByte_waitSibTx
; Wait until all SIB bytes were received:
loop_readByte_waitSibRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_readByte_waitSibRx
; updiUartDataBuf_ram[10] now contains the ASCII-encoded NVM version number.
	ldiz	2*szNvmVersion
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram+10
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Give the target MCU a bit of time:
	rcall	updi_wait25ms
; Branch according to NVM version:
	cpi		tmp0, '0'
	breq	loop_readByte_nvm0
; NVM version number != 0, e.g. AVR_DB and AVR_DD.
; These devices apparently need a more complicated procedure:
	rcall	applyReset
	rcall	updi_idle
	rcall	updi_wait25ms
	rcall	releaseReset
	rcall	updi_idle
	rcall	updi_wait25ms
	rcall	sendNvmProgKey
loop_readByte_nvm0:
	rcall	applyReset
	rcall	updi_idle
	rcall	updi_wait25ms
	rcall	releaseReset
	rcall	updi_idle
	rcall	updi_wait25ms
; Send LDS instruction:
	rcall	sendLds
; Read answer byte from Target UPDI:
loop_readByte_waitLds:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_readByte_waitLds
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
	ldiz	2*szLds
	rcall	uart_putSz
    rcall   uart_putHexByte
    ldiz    2*szLineBreak
    rcall   uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop
	
loop_chipErase:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command, to keep UPDI alive and get the NVM version:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_chipErase_waitSibTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_chipErase_waitSibTx
; Wait until all SIB bytes were received:
loop_chipErase_waitSibRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_chipErase_waitSibRx
; updiUartDataBuf_ram[10] now contains the ASCII-encoded NVM version number.
	ldiz	2*szNvmVersion
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram+10
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	mov		tmp1, tmp0
; Enter the Chip Erase key:
	ldiz	2*szEnteringCEK
	rcall	uart_putSz
	rcall	sendChipEraseKey
	cpi		tmp1, '0'
	breq	loop_chipErase_confirmKey
; Enter the NVMPROG key, too (recommended in AVR64DD28 datasheet):
	ldiz	2*szEnteringNVMPK
	rcall	uart_putSz
	rcall	sendNvmProgKey
loop_chipErase_confirmKey:
; == Confirm success by reading ASI Key Status Register ==
	ldi		tmp0, UPDI_ASI_KEY_STATUS
	rcall	sendLdcs
; Read answer from target UPDI:
loop_chipErase_waitKeyStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_chipErase_waitKeyStatus
	ldiz	2*szAsiKeyStatus
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	bst		tmp0, UPDI_CHIPERASE_bp
	brts	loop_chipErase_keySuccess
; CHIPERASE key transmitting failed.
	ldiz	2*szErrKey
	rcall	uart_putSz
	rjmp	loop_exit
loop_chipErase_keySuccess:
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
	rcall	applyReset
	rcall	updi_idle
	rcall	releaseReset
	rcall	updi_idle
loop_chipErase_readLockStatus:
; Read LOCKSTATUS:
	rcall	updi_wait25ms
	ldi		tmp0, UPDI_ASI_SYS_STATUS
	rcall	sendLdcs
; Read answer from target UPDI:
loop_chipErase_waitLockStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_chipErase_waitLockStatus
	lds		tmp0, updiUartDataBuf_ram
	sbrc	tmp0, UPDI_LOCKSTATUS_bp
	rjmp	loop_chipErase_readLockStatus
	ldiz	2*szAsiSysStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	cpi		tmp1, '0'
	breq	loop_chipErase_end
; NVM version '2' provides an ERASEFAIL bit in ASI_SYS_STATUS.
	sbrs	tmp0, UPDI_ERASEFAIL_bp
	rjmp	loop_exit
	ldiz	2*szErrErasefail
	rcall	uart_putSz
loop_chipErase_end:
	rjmp	loop_exit

loop_writeFuse:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; === Get address from user ===
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte
	sts		writeFuseAddrH_ram, tmp0
	rcall	uart_getHexByte
	sts		writeFuseAddrL_ram, tmp0
; === Get data from user ===
	ldiz	2*szData
	rcall	uart_putSz
	rcall	uart_getHexByte
	sts		writeFuseData_ram, tmp0
	ldiz	2*szConfirm
	rcall	uart_putSz
	rcall	uart_getChar
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	cpi		tmp0, 'y'
	breq	loop_writeFuse_confirmed
; Operation cancelled by user.
	rjmp	loop
loop_writeFuse_confirmed:
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command, to keep UPDI alive and get the NVM version:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_writeFuse_waitSibTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_writeFuse_waitSibTx
; Wait until all SIB bytes were received:
loop_writeFuse_waitSibRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_writeFuse_waitSibRx
; updiUartDataBuf_ram[10] now contains the ASCII-encoded NVM version number.
	ldiz	2*szNvmVersion
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram+10
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	mov		tmp1, tmp0
; === Enter NVM programming mode (as done by 'pymcuprog') ===
	ldiz	2*szEnteringNVMPK
	rcall	uart_putSz
; Hold target device in reset:
	rcall	applyReset
	rcall	sendNvmProgKey
; Confirm key success by reading ASI Key Status Register:
	ldi		tmp0, UPDI_ASI_KEY_STATUS
	rcall	sendLdcs
; Read answer from target UPDI:
loop_writeFuse_waitKeyStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_writeFuse_waitKeyStatus
	ldiz	2*szAsiKeyStatus
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	bst		tmp0, UPDI_NVMPROG_bp
	brts	loop_writeFuse_nvmKeySuccess
; NVMPROG key transmitting failed.
	ldiz	2*szErrKey
	rcall	uart_putSz
	rcall	releaseReset
	rjmp	loop_exit
loop_writeFuse_nvmKeySuccess:
; Release reset and wait for unlock:
	rcall	applyReset
	rcall	updi_wait25ms
	rcall	releaseReset
; == Wait until PROGSTART bit is 1 ==
loop_writeFuse_progStartLoop:
	ldi		tmp0, UPDI_ASI_SYS_STATUS
	rcall	sendLdcs
; Read answer from target UPDI:
loop_writeFuse_waitSysStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_writeFuse_waitSysStatus
	lds		tmp0, updiUartDataBuf_ram
	sbrs	tmp0, UPDI_PROGSTART_bp
	rjmp	loop_writeFuse_progStartLoop
	ldiz	2*szAsiSysStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Do the NVM operation ==
	ldiz	2*szSendingSts0
	rcall	uart_putSz
; Write address LSB:
	rcall	updi_wait25ms
	ldi		tmp0, low(NVMCTRL_ADDR)
	ldi		tmp1, high(NVMCTRL_ADDR)
	lds		tmp2, writeFuseAddrL_ram
	rcall	sendSts
	cpi		tmp3, 0
	brne	loop_writeFuse_sts01
	ldiz	2*szOK
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop_writeFuse_sts1
loop_writeFuse_sts01:
	cpi		tmp3, 1
	brne	loop_writeFuse_sts02
	rjmp	loop_timeout
loop_writeFuse_sts02:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrInvAnswer
	rcall	uart_putSz
	rjmp	loop_exit
loop_writeFuse_sts1:
	ldiz	2*szSendingSts1
	rcall	uart_putSz
; Write address MSB:
	rcall	updi_wait25ms
	ldi		tmp0, low(NVMCTRL_ADDR+1)
	ldi		tmp1, high(NVMCTRL_ADDR+1)
	lds		tmp2, writeFuseAddrH_ram
	rcall	sendSts
	cpi		tmp3, 0
	brne	loop_writeFuse_sts11
	ldiz	2*szOK
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop_writeFuse_sts2
loop_writeFuse_sts11:
	cpi		tmp3, 1
	brne	loop_writeFuse_sts12
	rjmp	loop_timeout
loop_writeFuse_sts12:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrInvAnswer
	rcall	uart_putSz
	rjmp	loop_exit
loop_writeFuse_sts2:
	ldiz	2*szSendingSts2
	rcall	uart_putSz
; Write data:
	rcall	updi_wait25ms
	ldi		tmp0, low(NVMCTRL_DATA)
	ldi		tmp1, high(NVMCTRL_DATA)
	lds		tmp2, writeFuseData_ram
	rcall	sendSts
	cpi		tmp3, 0
	brne	loop_writeFuse_sts21
	ldiz	2*szOK
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop_writeFuse_sts3
loop_writeFuse_sts21:
	cpi		tmp3, 1
	brne	loop_writeFuse_sts22
	rjmp	loop_timeout
loop_writeFuse_sts22:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrInvAnswer
	rcall	uart_putSz
	rjmp	loop_exit
loop_writeFuse_sts3:
	ldiz	2*szSendingSts3
	rcall	uart_putSz
; Issue Write Fuse command:
	rcall	updi_wait25ms
	ldi		tmp0, low(NVMCTRL_CTRLA)
	ldi		tmp1, high(NVMCTRL_CTRLA)
	ldi		tmp2, NVMCTRL_CMD_FUSEWRITE_gc
	rcall	sendSts
	cpi		tmp3, 0
	brne	loop_writeFuse_sts31
	ldiz	2*szOK
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop_writeFuse_waitNvm
loop_writeFuse_sts31:
	cpi		tmp3, 1
	brne	loop_writeFuse_sts32
	rjmp	loop_timeout
loop_writeFuse_sts32:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrInvAnswer
	rcall	uart_putSz
	rjmp	loop_exit
loop_writeFuse_waitNvm:
; == Wait while NVMCTRL is busy ==
	rcall	waitForNvm
	sbrc	tmp3, 0
	rjmp	loop_timeout
	ldiz	2*szNvmctrlStatus
	rcall	uart_putSz
	mov		tmp0, tmp2
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
	rcall	applyReset
	rcall	updi_idle
	rcall	updi_wait25ms
	rcall	releaseReset
	rcall	updi_idle
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop_exit

loop_timeout:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrTimeout
	rcall	uart_putSz
; Send BREAK to re-establish synchronization and disable UPDI:
	rcall	updi_break
loop_exit:
	rcall	updi_disable
	rjmp	loop
	
; == Subroutines ==
; === Routine: Send character via UART ===
; Parameter: tmp0.
uart_putChar:
	push	tmp0
; Make sure the transmit data registers are empty:
uart_sendByte_wait:
	xin		tmp0, USART0_STATUS
	sbrs	tmp0, USART_DREIF_bp
	rjmp	uart_sendByte_wait
	pop		tmp0
	xout	USART0_TXDATAL, tmp0
	ret

; === Routine: Send string via UART ===
; Parameter: Z pointer, start address of the null-terminated string.
; String length is limited to 255 characters.
uart_putSz:
	push	tmp0
	push	tmp1
	clr		tmp1
uart_putSz_loop:
	lpm		tmp0, Z+
	tst		tmp0
	breq	uart_putSz_end
	rcall	uart_putChar
	inc		tmp1
	cpi		tmp1, 0xFF
	breq	uart_putSz_end
	rjmp	uart_putSz_loop
uart_putSz_end:
	pop		tmp1
	pop		tmp0
	ret

; === Routine: Send byte in hexadecimal via UART ===
; Send 8-bit unsigned number as two hex digits.
; Parameter: tmp0.
uart_putHexByte:
	push	tmp0
	swap	tmp0
	andi	tmp0, 0x0F
	rcall	uart_putHexDigit
	pop		tmp0
	push	tmp0
	andi	tmp0, 0x0F
	rcall	uart_putHexDigit
	pop		tmp0
	ret
uart_putHexDigit:
	cpi		tmp0, 10
	brlt	uart_putHexDigit_1
	subi	tmp0, -('A'-'9'-1)
uart_putHexDigit_1:
	subi	tmp0, -'0'
	rcall	uart_putChar
	ret

; === Wait for reception of character from UART ===
; Return: tmp0, the received character.
uart_getChar:
	xin		tmp0, USART0_STATUS
	sbrs	tmp0, USART_RXCIF_bp
	rjmp	uart_getChar
	xin		tmp0, USART0_RXDATAL
	ret
	
; === Convert ASCII hex digit character to its value ===
; Parameter:	tmp0, the hex digit (ASCII) to be converted.
; Return:		tmp1, $FF if given character is no hex digit.
hexChar2num:
	mov		tmp1, tmp0
; Convert small letters to capitals:
	sbrc	tmp1, 6
	andi	tmp1, ~(1<<5)
	cpi		tmp1, '0'
	brlo	hexChar2num_err
	cpi		tmp1, '9'+1
	brlo	hexChar2num_09
	cpi		tmp1, 'A'
	brlo	hexChar2num_err
	cpi		tmp1, 'F'+1
	brlo	hexChar2num_AF
hexChar2num_err:
	ldi		tmp1, 0xFF
	rjmp	hexChar2num_end
hexChar2num_09:
	subi	tmp1, '0'
	rjmp	hexChar2num_end
hexChar2num_AF:
	subi	tmp1, 'A'-10
hexChar2num_end:
	ret

; == Read byte as two hex chars from UART ==
; Return:		tmp0.
uart_getHexByte:
	push	tmp1
	push	tmp2
uart_getHexByte_get1stChar:
	rcall	uart_getChar
	rcall	hexChar2num
	cpi		tmp1, 0xFF
	breq	uart_getHexByte_get1stChar
	rcall	uart_putChar				; Echo the first entered character.
	mov		tmp2, tmp1
uart_getHexByte_get2ndChar:
	rcall	uart_getChar
	rcall	hexChar2num
	cpi		tmp1, 0xFF
	breq	uart_getHexByte_get2ndChar
	rcall	uart_putChar				; Echo the second entered character.
	swap	tmp2
	or		tmp1, tmp2
	mov		tmp0, tmp1
	pop		tmp2
	pop		tmp1
	ret

; == Send the NVMPROG key ==
sendNvmProgKey:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendNvmProgKey_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendNvmProgKey_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp0, UPDI_SYNCH
	st		Z+, tmp0
	ldi		tmp0, 0b111_00_0_00	; Send 64-bit key.
	st		Z+, tmp0
	ldi		tmp0, 0x20				; Key byte 0.
	st		Z+, tmp0
	ldi		tmp0, 0x67				; Key byte 1.
	st		Z+, tmp0
	ldi		tmp0, 0x6F				; Key byte 2.
	st		Z+, tmp0
	ldi		tmp0, 0x72				; Key byte 3.
	st		Z+, tmp0
	ldi		tmp0, 0x50				; Key byte 4.
	st		Z+, tmp0
	ldi		tmp0, 0x4D				; Key byte 5.
	st		Z+, tmp0
	ldi		tmp0, 0x56				; Key byte 6.
	st		Z+, tmp0
	ldi		tmp0, 0x4E				; Key byte 7.
	st		Z+, tmp0
	ldi		tmp0, 10	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, tmp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Send the Chip Erase key ==
sendChipEraseKey:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendChipEraseKey_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendChipEraseKey_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp0, UPDI_SYNCH
	st		Z+, tmp0
	ldi		tmp0, 0b111_00_0_00	; Send 64-bit key.
	st		Z+, tmp0
	ldi		tmp0, 0x65				; Key byte 0.
	st		Z+, tmp0
	ldi		tmp0, 0x73				; Key byte 1.
	st		Z+, tmp0
	ldi		tmp0, 0x61				; Key byte 2.
	st		Z+, tmp0
	ldi		tmp0, 0x72				; Key byte 3.
	st		Z+, tmp0
	ldi		tmp0, 0x45				; Key byte 4.
	st		Z+, tmp0
	ldi		tmp0, 0x4D				; Key byte 5.
	st		Z+, tmp0
	ldi		tmp0, 0x56				; Key byte 6.
	st		Z+, tmp0
	ldi		tmp0, 0x4E				; Key byte 7.
	st		Z+, tmp0
	ldi		tmp0, 10	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, tmp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Apply reset to the target device ==
applyReset:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
applyReset_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	applyReset_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp0, UPDI_SYNCH
	st		Z+, tmp0
	ldi		tmp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	st		Z+, tmp0
	ldi		tmp0, 0x59					; Normal Reset signature.
	st		Z+, tmp0
	ldi		tmp0, 3					; Count of bytes to be sent.
	sts		updiUartDataCount_ram, tmp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret
	
; == Release reset from the target device ==
releaseReset:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
releaseReset_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	releaseReset_wait
; Clear the system reset:
	ldiz	updiUartDataBuf_ram
	ldi		tmp0, UPDI_SYNCH
	st		Z+, tmp0
	ldi		tmp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	st		Z+, tmp0
	ldi		tmp0, 0x00					; Clear reset.
	st		Z+, tmp0
	ldi		tmp0, 3					; Count of bytes to be sent.
	sts		updiUartDataCount_ram, tmp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Send LDCS instruction ==
; Parameters:	tmp0, the UPDI register to be loaded.
;				Must be in range 0x00..0x0F.
sendLdcs:
	push	tmp1
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendLdcs_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendLdcs_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp1, UPDI_SYNCH
	st		Z+, tmp1
	ldi		tmp1, 0b100_0_0000				; LDCS instruction.
	or		tmp1, tmp0
	st		Z+, tmp1			
	ldi		tmp1, 2
	sts		updiUartDataCount_ram, tmp1
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp1
	ret

; == Send LDS instruction ==
; Parameters:	YL, address LSB,
;				YH, address MSB.
; Original version:
sendLds:
	push	tmp2
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendLds_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendLds_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp2, UPDI_SYNCH
	st		Z+, tmp2
	ldi		tmp2, 0b000_0_01_00			; LDS instruction (word address, 1 byte of data).
	st		Z+, tmp2
	st		Z+, YL							; Address LSB.
	st		Z+, YH							; Address MSB.
	ldi		tmp2, 4
	sts		updiUartDataCount_ram, tmp2
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp2
	ret

; == Send STS instruction ==
; Parameters:	tmp0, address LSB,
;				tmp1, address MSB,
;				tmp2, data.
; Return:		tmp3, 0: success, 1: timeout, 2: invalid answer.
sendSts:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendSts_waitBusy:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendSts_waitBusy
; Send instruction code and address:
	ldiz	updiUartDataBuf_ram
	ldi		tmp3, UPDI_SYNCH
	st		Z+, tmp3						; SYNCH byte.
	ldi		tmp3, 0b010_0_01_00
	st		Z+, tmp3						; STS instruction (word address, 1 byte of data).
	st		Z+, tmp0						; Address LSB.
	st		Z+, tmp1						; Address MSB.
	ldi		tmp3, 4
	sts		updiUartDataCount_ram, tmp3	; Transfer block size.
	rcall	updi_tx
; Wait for first 'ACK' from Target UPDI:
sendSts_wait0:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	sendSts_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	sendSts_wait0
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
    cpi		tmp0, UPDI_ACK
	brne	sendSts_invAnswer
; Send data byte:
	sts		updiUartDataBuf_ram, tmp2		; Data byte to be written.
	ldi		tmp3, 1
	sts		updiUartDataCount_ram, tmp3	; Transfer block size.
	rcall	updi_tx
; Wait for second 'ACK' from Target UPDI:
sendSts_wait1:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	sendSts_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	sendSts_wait1
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
    cpi		tmp0, UPDI_ACK
    brne	sendSts_invAnswer
; Success.
	ldi		tmp3, 0
	rjmp	sendSts_pop
sendSts_invAnswer:
	ldi		tmp3, 2
	rjmp	sendSts_pop
sendSts_timeout:
	ldi		tmp3, 1
sendSts_pop:
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Wait while NVMCTRL is busy ==
; Return:		tmp2, last value of NVMCTRL_STATUS.
;				tmp3, 0x00: success, 0x01: timeout.
waitForNvm:
	push	tmp0
	push	tmp1
waitForNvm_loop:
	rcall	updi_idle
	rcall	updi_wait25ms
	ldi		tmp0, low(NVMCTRL_STATUS)
	ldi		tmp1, high(NVMCTRL_STATUS)
	rcall	sendLds
; Read answer byte from Target UPDI:
waitForNvm_waitRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	waitForNvm_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	waitForNvm_waitRx
; updiUartData_ram now contains the received byte.
    lds     tmp2, updiUartDataBuf_ram
	mov		tmp1, tmp2
	andi	tmp1, NVMCTRL_EEBUSY_bm | NVMCTRL_FBUSY_bm
	brne	waitForNvm_loop		; Branch if EEBUSY or FBUSY was set.
	ldi		tmp3, 0
	rjmp	waitForNvm_pop
waitForNvm_timeout:
	ldi		tmp3, 1
waitForNvm_pop:
	pop		tmp1
	pop		tmp0
	ret
	