; File:		avr-asm-lib\updi.asm
; Device:	ATtiny3216
; Created:  2024-06-14
; Version:	2024-06-20
; Author:	Johannes Fechner

; Special UPDI bytes:
.equ UPDI_SYNCH = 0x55	; Synchronization byte.
.equ UPDI_ACK = 0x40	; Acknowledge (= "successfully written, ready to accept new transmission" answer) byte.

; UPDI instructions and arguments:
.equ UPDI_KEY = 0xE0
.equ UPDI_KEY_64BITS = 0x00
.equ UPDI_KEY_128BITS = 0x01
.equ UPDI_KEY_GET_SIB = 1<<2
.equ UPDI_KEY_SEND_KEY = 0x00

; UPDI register addresses and bit positions:
.equ UPDI_ASI_KEY_STATUS = 0x07
.equ UPDI_CHIPERASE_bp = 3
.equ UPDI_NVMPROG_bp = 4
.equ UPDI_ASI_SYS_STATUS = 0x0B
.equ UPDI_LOCKSTATUS_bp = 0
.equ UPDI_PROGSTART_bp = 3 ; also called 'NVMPROG'.
.equ UPDI_ERASEFAIL_bp = 6

; Other UPDI properties:
.equ UPDI_SIB_LENGTH = 16 ; System Information Block length in bytes.

; Calculations:
.equ UPDI_TIMER_UART_PERIOD = ROUND(F_CPU/(1.0*UPDI_BPS))
.equ UPDI_TIMER_ENABLE_TIMEOUT_PERIOD = ROUND(1.0e-3*F_CPU)		; 1ms.
.equ UPDI_TIMER_UART_TIMEOUT_PERIOD = ROUND(1.0*F_CPU/1024.0)	; 1s.

; UPDI configuration and options:
.equ UPDI_RX_TIMEOUT = 5 ; seconds.
.equ UPDI_TX_USE_IDLE_BITS = 0 ; Whether to insert 2 idle bits between transmitted bytes.

; == Routines ==
; === Enable Target UPDI ===
updi_enable:
	push	tmp0
; Disable Target UPDI falling-edge interrupt:
	xin		tmp0, TARGET_UPDI_PINCTRL
	andi	tmp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, tmp0
; === Imitate start sequence of mEDBG ===
; Pull Target UPDI low for c. 50ms:
/*	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_wait25ms
	rcall	updi_wait25ms
; Release and wait for c. 100ms:
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_wait25ms
	rcall	updi_wait25ms
	rcall	updi_wait25ms
	rcall	updi_wait25ms
; Pull Target UPDI low, once again for c. 50ms:
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_wait25ms
	rcall	updi_wait25ms
; Release Target UPDI:
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp */
; === "Actual" Target UPDI Enable procedure ===
; Pull Target UPDI low for c. 500ns:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
.equ DELAY_LOOP_CYCLES = ROUND(((F_CPU*3.0e-7)-3.0)/2.0)
.if DELAY_LOOP_CYCLES < 1
.error "updi_enable delayLoop has to be replaced by a couple of 'nop' instructions."
.endif
	ldi		tmp0, DELAY_LOOP_CYCLES
updi_enable_delayLoop:
	dec		tmp0
	brne	updi_enable_delayLoop
; Release Target UPDI:
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
; Wait for target device to get ready:
; Configure timer for time measuring and timeout detection:
/*	ldi		tmp0, low(UPDI_TIMER_ENABLE_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER, tmp0
	ldi		tmp0, high(UPDI_TIMER_ENABLE_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER+1, tmp0
	ldi		tmp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, tmp0
	ldi		tmp0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, tmp0*/
updi_enable_waitForTarget:
/*	bst		updiFlags, UPDI_FLAGS_TIMEOUT_bp
	brts	updi_enable_afterWait*/
	sbis	TARGET_UPDI_VIN, TARGET_UPDI_bp
	rjmp	updi_enable_waitForTarget
updi_enable_afterWait:
; Stop timer:
/*	xout	TCA0_SINGLE_CTRLA, zero*/
;	rcall	updi_idle
	pop		tmp0
	ret

; === Disable Target UPDI ===
updi_disable:
	push	tmp0
; Force guard time delay:
	rcall	updi_idle
	ldi		tmp0, UPDI_SYNCH
	sts		updiUartDataBuf_ram, tmp0
	ldi		tmp0, 0b110_0_0000 | 0x03	; UPDI_CTRLB.
	sts     updiUartDataBuf_ram+1, tmp0
	ldi		tmp0, 1<<2					; UPDIDIS bit.
	sts     updiUartDataBuf_ram+2, tmp0
	ldi		tmp0, 3
	sts     updiUartDataCount_ram, tmp0
	rcall   updi_tx
	pop		tmp0
	ret

; === Wait for c. 25ms (for BREAK) ===
.equ UPDI_BREAK_CPU_CYCLES = ROUND(0.025*F_CPU/10.0)
.if UPDI_BREAK_CPU_CYCLES > 0xFFFF
.error "updi_wait25ms needs a third counter register."
.endif
updi_wait25ms:
	push	tmpL
	push	tmpH
	ldi		tmpL, low(UPDI_BREAK_CPU_CYCLES)
	ldi		tmpH, high(UPDI_BREAK_CPU_CYCLES)
updi_wait25ms_loop:
	nop
	nop
	nop
	nop
	nop
	nop
	sbiw	tmpH:tmpL, 1
	brne	updi_wait25ms_loop
	pop		tmpH
	pop		tmpL
	ret

; === Send two BREAKs ===
; Duration of a BREAK and the pause between two BREAKs:
updi_break:
; Wait for any previous transmission or reception to be completed:
updi_break_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_break_wait
; Send first BREAK:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_wait25ms
; Pause:
	sbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rcall	updi_wait25ms
; Send second BREAK:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rcall	updi_wait25ms
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_idle
	ret

; === Transmit the byte(s) stored in updiUartDataBuf_ram via UPDI ===
; Parameters:   updiUartDataBuf_ram, the byte(s) to be transmitted (including
;               the preceding SYNCH byte, if necessary); sent in ascending address order.
;               updiUartDataCount_ram, the count of bytes stored in updiUartDataBuf_ram
;               (ranging from 1 to the size of updiUartDataBuf_ram).
updi_tx:
	push	tmp0
; Wait for any previous transmission or reception to be completed:
updi_tx_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_tx_wait
; Disable interrupts:
	cli
; Stop UPDI UART timer (previously used for timeout detection):
	xout	TCA0_SINGLE_CTRLA, zero
; Cancel potential pending timer interrupt:
	xin		tmp0, TCA0_SINGLE_INTFLAGS
	ori		tmp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, tmp0
; Clear TXC, RXC, and timeout flags:
	mov		tmp0, updiFlags
	andi	tmp0, ~((1<<UPDI_FLAGS_TXC_bp)|(1<<UPDI_FLAGS_RXC_bp)|(1<<UPDI_FLAGS_TIMEOUT_bp))
	mov		updiFlags, tmp0
	sei
; Set busy flag:
	set
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
; Initialize bit counter and data byte index:
	sts		updiUartBitCntr_ram, zero
	sts		updiUartDataIndex_ram, zero
; Disable Target UPDI edge interrupt:
	xin		tmp0, TARGET_UPDI_PINCTRL
	andi	tmp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, tmp0
; Set Target UPDI pin as low-level output:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
; Configure, reset and start UPDI UART timer:
	ldi		tmp0, low(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER, tmp0
	ldi		tmp0, high(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER+1, tmp0
	ldi		tmp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, tmp0
	ldi		tmp0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, tmp0
; Re-enable interrupts:
	sei
; Switch on UPDI Tx LED:
	sbi		LED_UPDI_TX_VOUT, LED_UPDI_TX_bp
	pop		tmp0
	ret

; === Insert IDLE frame ===
; Wait until current transmission or reception is completed and pause for
; approx. one UPDI UART frame duration.
updi_idle:
	push	tmpL
	push	tmpH
; Wait for any previous transmission or reception to be completed:
updi_idle_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_idle_wait
; Pause for approx. one UPDI UART bit duration:
.equ UPDI_IDLE_CPU_CYCLES = ROUND(12.0*F_CPU/(UPDI_BPS*10.0))
	ldi		tmpL, low(UPDI_IDLE_CPU_CYCLES)
	ldi		tmpH, high(UPDI_IDLE_CPU_CYCLES)
updi_idle_delay:
	nop
	nop
	nop
	nop
	nop
	nop
	sbiw	tmpL, 1
	brne	updi_idle_delay
	pop		tmpH
	pop		tmpL
	ret

; === Append byte to buffer ===
; Parameter:	tmp0, the byte to be appended.
updi_appendBuf:
	push	tmp1
	push	ZL
	push	ZH
	lds		tmp1, updiUartDataCount_ram
	ldiz	updiUartDataBuf_ram
	addz	tmp1
	st		Z, tmp0
	inc		tmp1
	sts		updiUartDataCount_ram, tmp1
	pop		ZH
	pop		ZL
	pop		tmp1
	ret

; == Interrupt Service Routines ==
; === UPDI Falling-Edge ISR ===
updi_fallingEdgeIsr:
	push	tmp0
	xin		tmp0, CPU_SREG
	push	tmp0
; Stop UPDI UART timer (previously used for timeout detection):
	xout	TCA0_SINGLE_CTRLA, zero
; Clear interrupt flag:
	xin		tmp0, TARGET_UPDI_INTFLAGS
	ori		tmp0, (1<<TARGET_UPDI_bp)
	xout	TARGET_UPDI_INTFLAGS, tmp0
; Clear timer interrupt flag, too:
	xin		tmp0, TCA0_SINGLE_INTFLAGS
	ori		tmp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, tmp0
; Set busy flag:
	set
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
; Disable Target UPDI edge interrupt:
	xin		tmp0, TARGET_UPDI_PINCTRL
	andi	tmp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, tmp0
; Initialize bit counter:
	sts		updiUartBitCntr_ram, zero
; Configure UPDI UART timer:
	ldi		tmp0, low(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER, tmp0
	ldi		tmp0, high(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER+1, tmp0
; Load the counter register with half the period so that the received frame is
; sampled in the middle of each bit duration:
	ldi		tmp0, low(ROUND(UPDI_TIMER_UART_PERIOD/2.0))
	xout	TCA0_SINGLE_CNT, tmp0
	ldi		tmp0, high(ROUND(UPDI_TIMER_UART_PERIOD/2.0))
	xout	TCA0_SINGLE_CNT+1, tmp0
; Start the timer:
	ldi		tmp0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, tmp0
; Switch on UPDI Rx LED:
	sbi		LED_UPDI_RX_VOUT, LED_UPDI_RX_bp
updi_fallingEdgeIsr_pop:
	pop		tmp0
	xout	CPU_SREG, tmp0
	pop		tmp0
	reti

; === UPDI Timer ISR ===
updi_timerIsr:
	push	tmp0
	push	tmp1
	push	ZL
	push	ZH
	xin		tmp0, CPU_SREG
	push	tmp0
; Clear interrupt flag:
	xin		tmp0, TCA0_SINGLE_INTFLAGS
	ori		tmp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, tmp0
; Busy or timer overflow while waiting for reception?
	sbrs    updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp    updi_timerIsr_rxTimeout
; Busy. Increment bit counter and write back:
	lds		tmp0, updiUartBitCntr_ram
	inc		tmp0
	sts		updiUartBitCntr_ram, tmp0
; Transmitting or receiving?
	sbis	TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rjmp	updi_timerIsr_rx					; Target UPDI set as input -> receiving.
; Target UPDI set as output -> currently transmitting.
; tmp0 still contains bit counter value.
	cpi		tmp0,  1
	breq	updi_timerIsr_txAfterStartBit
	cpi		tmp0,  9
	breq	updi_timerIsr_txAfterLastDataBit
	cpi		tmp0, 10
	breq	updi_timerIsr_txAfterParityBit
	cpi		tmp0, 11
	breq	updi_timerIsr_txAfterParityBit
	cpi		tmp0, 12
	breq	updi_timerIsr_txEndOfFrame
.if UPDI_TX_USE_IDLE_BITS
	cpi		tmp0, 13
	breq	updi_timerIsr_txIdle0
	cpi		tmp0, 14
	brsh	updi_timerIsr_txIdle1
.endif
; (2 <= bit counter <= 8) -> [2nd to 8th] data bit follows.
updi_timerIsr_txInsideData:
; Load data byte and bit mask, set Target UPDI output accordingly:
	ldiz	updiUartDataBuf_ram
	lds		tmp0, updiUartDataIndex_ram
	addz	tmp0
	ld		tmp0, Z
	lds		tmp1, updiUartBitMask_ram
	and		tmp0, tmp1
	breq	updi_timerIsr_txInsideData_0		; Branch if data bit is 0.
	sbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	lds		tmp0, updiUartParityCntr_ram
	inc		tmp0
	sts		updiUartParityCntr_ram, tmp0
	rjmp	updi_timerIsr_txShiftBitMask
updi_timerIsr_txInsideData_0:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
updi_timerIsr_txShiftBitMask:
; Update bit mask and write back:
	lsl		tmp1
	sts		updiUartBitMask_ram, tmp1
	rjmp	updi_timerIsr_pop
updi_timerIsr_txAfterStartBit:
; Initialize bit mask and parity counter:
	ldi		tmp1, 1
	sts		updiUartBitMask_ram, tmp1
	sts		updiUartParityCntr_ram, zero
	rjmp	updi_timerIsr_txInsideData
updi_timerIsr_txAfterLastDataBit:
; Set Target UPDI output according to parity:
	lds		tmp0, updiUartParityCntr_ram
	bst		tmp0, 0
	xin		tmp0, TARGET_UPDI_VOUT
	bld		tmp0, TARGET_UPDI_bp
	xout	TARGET_UPDI_VOUT, tmp0
; Increment the byte index:
	lds		tmp0, updiUartDataIndex_ram
	inc		tmp0
	sts		updiUartDataIndex_ram, tmp0
	rjmp	updi_timerIsr_pop
updi_timerIsr_txAfterParityBit:
; Stop bit (high) follows:
	sbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_txEndOfFrame:
; After 2nd=last stop bit -> end of frame.
; Check whether end of transmission is reached:
	lds		tmp0, updiUartDataIndex_ram
	lds		tmp1, updiUartDataCount_ram
	cp		tmp0, tmp1
	brsh	updi_timerIsr_txEot		; Branch if EOT reached.
.if UPDI_TX_USE_IDLE_BITS
; Else, insert inter-byte delay of 2 idle bits.
updi_timerIsr_txIdle0:
	rjmp	updi_timerIsr_pop
updi_timerIsr_txIdle1:
.endif
; Reset bit counter and drive UPDI low for start bit:
	sts		updiUartBitCntr_ram, zero
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_txEot:
; End of transmission reached.
; Stop timer:
	xout	TCA0_SINGLE_CTRLA, zero
; Set Target UPDI pin as input:
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
; Clear busy flag, set TXC flag:
	clt
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
	set
	bld		updiFlags, UPDI_FLAGS_TXC_bp
; Enable Target UPDI edge interrupt:
	xin		tmp0, TARGET_UPDI_PINCTRL
	ori		tmp0, PORT_ISC_FALLING_gc
	xout	TARGET_UPDI_PINCTRL, tmp0
; Clear data byte index:
	sts		updiUartDataIndex_ram, zero
; Configure timer for RX-timeout detection:
	sts		updiUartTimeoutCntr_ram, zero
	ldi		tmp0, low(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER, tmp0
	ldi		tmp0, high(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER+1, tmp0
	ldi		tmp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, tmp0
	ldi		tmp0, TCA_SINGLE_CLKSEL_DIV1024_gc | TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, tmp0
; Switch off UPDI Tx LED:
	cbi		LED_UPDI_TX_VOUT, LED_UPDI_TX_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_rxTimeout:
	lds		tmp0, updiUartTimeoutCntr_ram
	inc		tmp0
	sts		updiUartTimeoutCntr_ram, tmp0
	cpi		tmp0, UPDI_RX_TIMEOUT
	brlo	updi_timerIsr_noTimeout
; Timeout reached.
; Disable Target UPDI edge interrupt:
	xin		tmp0, TARGET_UPDI_PINCTRL
	andi	tmp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, tmp0
; Disable potential pending interrupt:
	xin		tmp0, TARGET_UPDI_INTFLAGS
	ori		tmp0, (1<<TARGET_UPDI_bp)
	xout	TARGET_UPDI_INTFLAGS, tmp0
; Stop timer:
	xout	TCA0_SINGLE_CTRLA, zero
; Set timeout flag:
	set
	bld     updiFlags, UPDI_FLAGS_TIMEOUT_bp
updi_timerIsr_noTimeout:
	rjmp    updi_timerIsr_pop
updi_timerIsr_rx:
; tmp0 still contains bit counter value.
	cpi		tmp0, 1
	breq	updi_timerIsr_rxStart
	cpi		tmp0, 10
	breq	updi_timerIsr_rxParity
	cpi		tmp0, 11
	breq	updi_timerIsr_rx1stStop
	cpi		tmp0, 12
	brsh	updi_timerIsr_rx2ndStop
; (2 <= bit counter <= 9) -> data bit.
	ldiz	updiUartDataBuf_ram
	lds		tmp0, updiUartDataIndex_ram
; Make sure index is in allowed range:
	cpi		tmp0, UPDI_UART_BUFLEN
	brsh	updi_timerIsr_bufovf
; Load the appropriate data byte from buffer:
	addz	tmp0
	ld		tmp0, Z
	lds		tmp1, updiUartBitMask_ram
; If UPDI is high, set the appropriate bit:
	sbic	TARGET_UPDI_VIN, TARGET_UPDI_bp
	or		tmp0, tmp1
	st		Z, tmp0
; Left-shift the bit mask and write back:
	lsl		tmp1
	sts		updiUartBitMask_ram, tmp1
	rjmp	updi_timerIsr_pop
updi_timerIsr_bufovf:
	set
	bld		updiFlags, UPDI_FLAGS_BUFOVF_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_rxStart:
; Initialize the bit mask and clear the data byte:
	ldi		tmp0, 1
	sts		updiUartBitMask_ram, tmp0
	ldiz	updiUartDataBuf_ram
	lds		tmp0, updiUartDataIndex_ram
; Make sure index is in allowed range:
	cpi		tmp0, UPDI_UART_BUFLEN
	brsh	updi_timerIsr_bufovf
	addz	tmp0
	st		Z, zero
	rjmp	updi_timerIsr_pop
updi_timerIsr_rxParity:
; ##### To be done: test for correct parity. #####
; Increment byte index:
	lds		tmp0, updiUartDataIndex_ram
	inc		tmp0
	sts		updiUartDataIndex_ram, tmp0
updi_timerIsr_rx1stStop:
; ##### To be done: test for high level. #####
	rjmp	updi_timerIsr_pop
updi_timerIsr_rx2ndStop:
; End of frame reached.
; ##### To be done: test for high level. #####
; Stop timer:
	xout	TCA0_SINGLE_CTRLA, zero
; Set UPDI Frame Received flag:
	set
	bld		updiFlags, UPDI_FLAGS_RXC_bp
; Clear busy flag:
    clt
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
; Enable Target UPDI edge interrupt:
	xin		tmp0, TARGET_UPDI_PINCTRL
	ori		tmp0, PORT_ISC_FALLING_gc
	xout	TARGET_UPDI_PINCTRL, tmp0
; Configure timer for RX-timeout detection:
	sts		updiUartTimeoutCntr_ram, zero
	ldi		tmp0, low(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER, tmp0
	ldi		tmp0, high(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER+1, tmp0
	ldi		tmp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, tmp0
	ldi		tmp0, TCA_SINGLE_CLKSEL_DIV1024_gc | TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, tmp0
; Switch off UPDI Rx LED:
	cbi		LED_UPDI_RX_VOUT, LED_UPDI_RX_bp
updi_timerIsr_pop:
	pop		tmp0
	xout	CPU_SREG, tmp0
	pop		ZH
	pop		ZL
	pop		tmp1
	pop		tmp0
	reti
